@@ -1,6 +1,10 @@
 /*
- * $Id: ngrep.c,v 1.37 2000/02/01 12:28:02 jpr5 Exp $
+ * $Id: ngrep.c,v 1.38 2000/06/26 19:52:50 jpr5 Exp $
  *
+ * Copyright (c) 2000  Jordan Ritter <jpr5@darkridge.com>
+ *
+ * Please refer to the COPYRIGHT file for more information. 
+ * 
  */
 
 #if defined(BSD) || defined(SOLARIS)
@@ -24,6 +28,8 @@
 
 #if defined(LINUX)
 #include <getopt.h>
+#include <arpa/inet.h>
+#include <ctype.h>
 #endif
 
 
@@ -43,25 +49,35 @@
 #include "ngrep.h"
 
 
-static char rcsver[] = "$Revision: 1.37 $";
+static char rcsver[] = "$Revision: 1.38 $";
 
 int snaplen = 65535, promisc = 1, to = 1000;
 int show_empty = 0, show_hex = 0, quiet = 0;
-int match_after = 0, keep_matching = 0, invert_match = 0;
+int match_after = 0, keep_matching = 0;
+int invert_match = 0, bin_match = 0;
 int matches = 0, max_matches = 0;
+int live_read = 1, want_delay = 0;;
 
 char pc_err[PCAP_ERRBUF_SIZE], *re_err;
 
 int re_match_word = 0, re_ignore_case = 0;
-char *regex = NULL, *filter = NULL;
 struct re_pattern_buffer pattern;
-int (*match_func)();
+
+char *match_data = NULL, *bin_data = NULL, *filter = NULL;
+int (*match_func)() = &blank_match_func;
+int match_len = 0;
 
 struct bpf_program pcapfilter;
 struct in_addr net, mask;
+pcap_t *pd = NULL;
 char *dev = NULL;
 int link_offset;
-pcap_t *pd;
+
+char *read_file = NULL, *dump_file = NULL;
+pcap_dumper_t *pd_dump = NULL;
+
+struct timeval prev_ts = {0, 0}, prev_delay_ts = {0,0};
+void (*print_time)() = NULL;
 
 
 int main(int argc, char **argv) {
@@ -72,8 +88,14 @@ int main(int argc, char **argv) {
   signal(SIGABRT, clean_exit);
   signal(SIGPIPE, clean_exit);
 
-  while ((c = getopt(argc, argv, "hViwqevxln:d:A:")) != EOF) {
+  while ((c = getopt(argc, argv, "hXViwqevxlDtTn:d:A:I:O:")) != EOF) {
     switch (c) {
+    case 'I':  
+      read_file = optarg;
+      break;
+    case 'O':
+      dump_file = optarg;
+      break;
     case 'A': 
       match_after = atoi(optarg) + 1;
       break;
@@ -83,6 +105,15 @@ int main(int argc, char **argv) {
     case 'n':
       max_matches = atoi(optarg);
       break;
+    case 'T':
+      print_time = &print_time_diff;
+      break;
+    case 't':
+      print_time = &print_time_absolute;
+      break;
+    case 'D':
+      want_delay = 1;
+      break;
     case 'l':
       setvbuf(stdout, NULL, _IOLBF, 0);
       break;
@@ -106,6 +137,9 @@ int main(int argc, char **argv) {
       break;
     case 'V':
       version();
+    case 'X':
+      bin_match++;
+      break;
     case 'h':
       usage(0);
     default:
@@ -113,35 +147,45 @@ int main(int argc, char **argv) {
     }
   }
 
+
   if (argv[optind]) 
-    regex = argv[optind++];
+    match_data = argv[optind++];
 
-
-  if (!dev) 
-    if (!(dev = pcap_lookupdev(pc_err))) {
+  if (read_file) {
+    if (!(pd = pcap_open_offline(read_file, pc_err))) {
       perror(pc_err);
-      exit(-1);
+      clean_exit(-1);
     }
 
+    live_read = 0;
+    printf("input: %s\n", read_file);
 
-  if ((pd = pcap_open_live(dev, snaplen, promisc, to, pc_err)) == NULL) {
-    perror(pc_err);
-    exit(-1);
-  }
+  } else {
+    if (!dev)
+      if (!(dev = pcap_lookupdev(pc_err))) {
+	perror(pc_err);
+	clean_exit(-1);
+      }
+    
+    if ((pd = pcap_open_live(dev, snaplen, promisc, to, pc_err)) == NULL) {
+      perror(pc_err);
+      clean_exit(-1);
+    }
 
-  if (pcap_lookupnet(dev, &net.s_addr, &mask.s_addr, pc_err) == -1) {
-    perror(pc_err);
-    memset(&net, 0, sizeof(net));
-    memset(&mask, 0, sizeof(mask));
-  } 
+    if (pcap_lookupnet(dev, &net.s_addr, &mask.s_addr, pc_err) == -1) {
+      perror(pc_err);
+      memset(&net, 0, sizeof(net));
+      memset(&mask, 0, sizeof(mask));
+    } 
 
-  if (!quiet) {
-    printf("interface: %s", dev);
-    if (net.s_addr && mask.s_addr) {
-      printf(" (%s/", inet_ntoa(net));
-      printf("%s)", inet_ntoa(mask)); 
+    if (!quiet) {
+      printf("interface: %s", dev);
+      if (net.s_addr && mask.s_addr) {
+	printf(" (%s/", inet_ntoa(net));
+	printf("%s)", inet_ntoa(mask)); 
+      }
+      printf("\n");
     }
-    printf("\n");
   }
 
 
@@ -155,64 +199,100 @@ int main(int argc, char **argv) {
       PCAP_RESTART();
       if (pcap_compile(pd, &pcapfilter, filter, 0, mask.s_addr)) {
 	pcap_perror(pd, "pcap compile");
-	exit(-1);
-      } else regex = NULL;
+	clean_exit(-1);
+      } else match_data = NULL;
     }
 
     if (!quiet) printf("filter: %s\n", filter); 
     
     if (pcap_setfilter(pd, &pcapfilter)) {
       pcap_perror(pd, "pcap set");
-      exit(-1);
+      clean_exit(-1);
     }
   }
 
+  if (match_data) {
+    if (bin_match) {
+      int i = 0, n;
+      char *s, *d;
+      int len;
 
-  if (regex) {
-    match_func = &re_match_func;
-    re_syntax_options = RE_SYNTAX_EGREP;
+      if (re_match_word || re_ignore_case) {
+	fprintf(stderr, "fatal: regex switches are incompatible with binary matching\n");
+	clean_exit(-1);
+      }
 
-    if (re_ignore_case) {
-      char *s;
-      int i;
+      len = strlen(match_data);
+      if (len % 2 != 0 || !strishex(match_data)) {
+	fprintf(stderr, "fatal: invalid hex string specified\n");
+	clean_exit(-1);
+      }
 
-      pattern.translate = (char*)malloc(256);
-      s = pattern.translate;
-      for (i = 0; i < 256; i++) 
-	s[i] = i;
-      for (i = 'A'; i <= 'Z'; i++) 
-	s[i] = i + 32;
+      bin_data = malloc(len / 2);
+      memset(bin_data, 0, len / 2);
+      d = bin_data;
 
-      s = regex;
-      while (*s) 
-	*s++ = tolower(*s);
-    } else pattern.translate = NULL;
+      if ((s = strchr(match_data, 'x'))) 
+	len -= ++s - match_data - 1;
+      else s = match_data;
 
-    if (re_match_word) {
-      char *word_regex = malloc(strlen(regex) * 3 + strlen(WORD_REGEX));
-      sprintf(word_regex, WORD_REGEX, regex, regex, regex);
-      regex = word_regex;
+      while (i <= len) {
+	sscanf(s+i, "%2x", &n);
+	*d++ = n;
+	i += 2;
+      }
+
+      match_len = len / 2;
+      match_func = &bin_match_func;
+
+    } else {
+      re_syntax_options = RE_SYNTAX_EGREP;
+      
+      if (re_ignore_case) {
+	char *s;
+	int i;
+	
+	pattern.translate = (char*)malloc(256);
+	s = pattern.translate;
+	for (i = 0; i < 256; i++) 
+	  s[i] = i;
+	for (i = 'A'; i <= 'Z'; i++) 
+	  s[i] = i + 32;
+
+	s = match_data;
+	while (*s) 
+	  *s++ = tolower(*s);
+      } else pattern.translate = NULL;
+      
+      if (re_match_word) {
+	char *word_regex = malloc(strlen(match_data) * 3 + strlen(WORD_REGEX));
+	sprintf(word_regex, WORD_REGEX, match_data, match_data, match_data);
+	match_data = word_regex;
+      }
+      
+      (const char *)re_err = re_compile_pattern(match_data, strlen(match_data), &pattern);
+      if (re_err) {
+	fprintf(stderr, "regex compile: %s\n", re_err);
+	clean_exit(-1);
+      }
+     
+      pattern.fastmap = (char*)malloc(256);
+      if (re_compile_fastmap(&pattern)) {
+	perror("fastmap compile failed");
+	clean_exit(-1);
+      }
+
+      match_func = &re_match_func;
     }
 
-    (const char *)re_err = re_compile_pattern(regex, strlen(regex), &pattern);
-    if (re_err) {
-      fprintf(stderr, "regex compile: %s\n", re_err);
-      exit(-1);
-    }
-
-    pattern.fastmap = (char*)malloc(256);
-    if (re_compile_fastmap(&pattern)) {
-      perror("fastmap compile failed");
-      exit(-1);
-    }
-
-    if (!quiet && regex && strlen(regex)) 
-      printf("%smatch: %s\n", invert_match?"don't ":"", regex);
-  } else match_func = &blank_match_func;
+    if (!quiet && match_data && strlen(match_data)) 
+      printf("%smatch: %s%s\n", invert_match?"don't ":"", 
+	     (bin_data && !strchr(match_data, 'x'))?"0x":"", match_data);
+  }
 
 
-  if (re_match_word) free(regex);
   if (filter) free(filter);
+  if (re_match_word) free(match_data);
 
 
   switch(pcap_datalink(pd)) {
@@ -243,11 +323,19 @@ int main(int argc, char **argv) {
 
   default:
     fprintf(stderr, "fatal: unsupported interface type\n");
-    exit(-1);
+    clean_exit(-1);
   }
   
+  if (dump_file) {
+    if (!(pd_dump = pcap_dump_open(pd, dump_file))) {
+      fprintf(stderr, "fatal: %s\n", pcap_geterr(pd));
+      clean_exit(-1);
+    } else printf("output: %s\n", dump_file);
+  }
 
   while (pcap_loop(pd, 0, (pcap_handler)process, 0));
+
+  clean_exit(0);
 }	
 
 
@@ -271,14 +359,21 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
       printf("#");
       fflush(stdout);
     }
-
+    
     data = ((char*)tcp) + tcphdr_offset;
     len = ntohs(ip_packet->ip_len) - ip_hl - tcphdr_offset;
 
     if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
 	|| keep_matching) { 
+
+      if (!live_read && want_delay)
+	dump_delay(h);
+
       printf("\nT ");
 
+      if (print_time) 
+	print_time(h);
+
       if (tcphdr_offset || !frag_offset) {
 	printf("%s:%d -", inet_ntoa(ip_packet->ip_src), ntohs(tcp->th_sport));
 	printf("> %s:%d", inet_ntoa(ip_packet->ip_dst), ntohs(tcp->th_dport));
@@ -294,12 +389,15 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
 	printf("> %s", inet_ntoa(ip_packet->ip_dst));
       }
 
-      if (fragmented) {
+      if (fragmented) 
 	printf(" %s%d@%d:%d\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
                                 frag_offset, len); 
-      } else printf("\n");
+      else printf("\n");
       
-      dump(data,len);
+      if (pd_dump) {
+	pcap_dump((u_char*)pd_dump, h, p);
+	if (!quiet) dump(data, len);
+      } else dump(data, len);
     }
   }
   break;
@@ -318,8 +416,15 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
 
     if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
 	|| keep_matching) { 
+
+      if (!live_read && want_delay)
+	dump_delay(h);
+
       printf("\nU ");
 
+      if (print_time) 
+	print_time(h);
+
       if (udphdr_offset || !frag_offset) {
 #ifdef HAVE_DUMB_UDPHDR
 	printf("%s:%d -", inet_ntoa(ip_packet->ip_src), ntohs(udp->source));
@@ -333,12 +438,15 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
 	printf("> %s", inet_ntoa(ip_packet->ip_dst));
       }
 
-      if (fragmented) {
+      if (fragmented) 
 	printf(" %s%d@%d:%d\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
-                                frag_offset, len); 
-      } else printf("\n");
+	       frag_offset, len); 
+      else printf("\n");
       
-      dump(data,len);
+      if (pd_dump) {
+	pcap_dump((u_char*)pd_dump, h, p);
+	if (!quiet) dump(data, len);
+      } else dump(data, len);
     }
   }
   break;
@@ -357,19 +465,30 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
 
     if (((len || show_empty) && (((int)(*match_func)(data, len)) != invert_match))
 	|| keep_matching) { 
+
+      if (!live_read && want_delay)
+	dump_delay(h);
+
       printf("\nI ");
 
+      if (print_time) 
+	print_time(h);
+
       printf("%s -", inet_ntoa(ip_packet->ip_src));
       printf("> %s", inet_ntoa(ip_packet->ip_dst));
 
       if (icmphdr_offset || !frag_offset) 
 	printf(" %d:%d", ic->icmp_type, ic->icmp_code);
 
-      if (fragmented) {
+      if (fragmented) 
 	printf(" %s%d@%d:%d\n", frag_offset?"+":"", ntohs(ip_packet->ip_id),
                                 frag_offset, len); 
-      } else printf("\n");
-      dump(data,len);
+      else printf("\n");
+
+      if (pd_dump) {
+	pcap_dump((u_char*)pd_dump, h, p);
+	if (!quiet) dump(data, len);
+      } else dump(data, len);
     }
   }
   break;
@@ -381,6 +500,55 @@ void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
 }
 
 
+int re_match_func(char *data, int len) {
+  switch (re_search(&pattern, data, len, 0, len, 0)) {
+    case -2: 
+      perror("she's dead, jim\n");
+      clean_exit(-2);
+    case -1:
+      return 0;
+  }
+  
+  if (max_matches && ++matches > max_matches)
+    clean_exit(0);
+
+  if (match_after && keep_matching != match_after)
+    keep_matching = match_after;
+
+  return 1;
+}
+
+
+int bin_match_func(char *data, int len) {
+  int stop = len - match_len;
+  int i = 0;
+
+  if (stop < 0)
+    return 0;
+
+  while (i <= stop) 
+    if (!memcmp(data+(i++), bin_data, match_len)) {
+      if (max_matches && ++matches > max_matches)
+	clean_exit(0);
+
+      if (match_after && keep_matching != match_after)
+	keep_matching = match_after;
+
+      return 1;
+    }
+
+  return 0;
+}
+
+
+int blank_match_func(char *data, int len) {
+  if (max_matches && ++matches > max_matches)
+    clean_exit(0);
+
+  return 1;
+}
+
+
 void dump(char *data, int len) {  
   if (len > 0) {
     unsigned width = show_hex?16:70;
@@ -396,7 +564,7 @@ void dump(char *data, int len) {
 	    printf("%02x ", (unsigned char)str[j]);
 	  else printf("   ");
 
-	  if((j+1) % (width/2) == 0)
+	  if ((j+1) % (width/2) == 0)
 	    printf("   ");
 	}
 
@@ -414,32 +582,6 @@ void dump(char *data, int len) {
 }
 
 
-int re_match_func(char *data, int len) {
-  switch (re_search(&pattern, data, len, 0, len, 0)) {
-    case -2: 
-      perror("she's dead, jim\n");
-      exit(-1);
-    case -1:
-      return 0;
-  }
-  
-  if (max_matches && ++matches > max_matches)
-    clean_exit(0);
-
-  if (match_after && keep_matching != match_after)
-    keep_matching = match_after;
-
-  return 1;
-}
-
-int blank_match_func(char *data, int len) {
-  if (max_matches && ++matches > max_matches)
-    clean_exit(0);
-
-  return 1;
-}
-
-
 char *get_filter(char **argv) {
   char **arg = argv, *theirs, *mine;
   char *from, *to;
@@ -466,7 +608,7 @@ char *get_filter(char **argv) {
     *(to-1) = ' ';
   }
 
-  sprintf(mine,IP_ONLY,theirs);
+  sprintf(mine, IP_ONLY, theirs);
 
   free(theirs);
   return mine;
@@ -474,16 +616,98 @@ char *get_filter(char **argv) {
 
 
 void clean_exit(int sig) {
+  struct pcap_stat s;
+  if (!quiet && sig >= 0) printf("exit\n");
+
   if (pattern.translate) free(pattern.translate);
   if (pattern.fastmap) free(pattern.fastmap);
 
-  printf("exit\n");
-  exit(0);
+  if (bin_data) free(bin_data);
+  
+  if (!quiet && sig >= 0 && !read_file &&
+      pd && !pcap_stats(pd, &s)) 
+    printf("%d received, %d dropped\n", s.ps_recv, s.ps_drop);
+
+  if (pd) pcap_close(pd);
+  if (pd_dump) pcap_dump_close(pd_dump);
+
+  exit(sig);
+}
+
+
+int strishex(char *str) {
+  char *s;
+  if ((s = strchr(str, 'x'))) 
+    s++;
+  else s = str;
+
+  while (*s) 
+    if (!isxdigit(*s++))
+      return 0;
+
+  return 1;
+}
+
+
+void print_time_absolute(struct pcap_pkthdr *h) {
+  struct tm *t = localtime(&h->ts.tv_sec);
+
+  printf("%02d/%02d/%02d %02d:%02d:%02d.%06d ",
+	 t->tm_year+1900, t->tm_mon, t->tm_mday, t->tm_hour,
+	 t->tm_min, t->tm_sec, h->ts.tv_usec);
+}
+
+
+void print_time_diff(struct pcap_pkthdr *h) { 
+  unsigned secs, usecs;
+
+  if (!prev_ts.tv_sec && !prev_ts.tv_usec) {// uninit
+    prev_ts.tv_sec = h->ts.tv_sec;
+    prev_ts.tv_usec = h->ts.tv_usec;
+  }
+
+  secs = h->ts.tv_sec - prev_ts.tv_sec;
+  if (h->ts.tv_usec >= prev_ts.tv_usec)
+    usecs = h->ts.tv_usec - prev_ts.tv_usec;
+  else {
+    secs--; 
+    usecs = 1000000 - (prev_ts.tv_usec - h->ts.tv_usec);
+  }
+
+  printf("+%d.%06d ", secs, usecs);
+
+  prev_ts.tv_sec = h->ts.tv_sec;
+  prev_ts.tv_usec = h->ts.tv_usec;
+}
+
+
+void dump_delay(struct pcap_pkthdr *h) {
+  unsigned secs, usecs;
+
+  if (!prev_delay_ts.tv_sec && !prev_delay_ts.tv_usec) {
+    prev_delay_ts.tv_sec = h->ts.tv_sec;
+    prev_delay_ts.tv_usec = h->ts.tv_usec;
+  }
+
+  secs = h->ts.tv_sec - prev_delay_ts.tv_sec;
+  if (h->ts.tv_usec >= prev_delay_ts.tv_usec)
+    usecs = h->ts.tv_usec - prev_delay_ts.tv_usec;
+  else {
+    secs--; 
+    usecs = 1000000 - (prev_delay_ts.tv_usec - h->ts.tv_usec);
+  }
+
+  sleep(secs);
+  usleep(usecs);
+  
+  prev_delay_ts.tv_sec = h->ts.tv_sec;
+  prev_delay_ts.tv_usec = h->ts.tv_usec;
 }
 
 
 void usage(int e) {
-  printf("usage: ngrep <-hViwqevxl> <-n num> <-d dev> <-A num> <regex> <pcap filter logic>\n");
+  printf("usage: ngrep <-hXViwqevxlDtT> <-IO pcap_dump> <-n num> <-d dev> <-A num>\n"
+	 "                              <match expression> <bpf filter>\n");
   exit(e);
 }
 
