@@ -1,10 +1,12 @@
 /*
- * $Id: ngrep.h,v 1.13 2000/02/01 12:28:08 jpr5 Exp $
+ * $Id: ngrep.h,v 1.14 2000/06/21 18:42:53 jpr5 Exp $
+ *
+ * Copyright (c) 2000  Jordan Ritter <jpr5@darkridge.com>
+ *
+ * Please refer to the COPYRIGHT file for more information. 
  *
  */
 
-/* for PPPHDR_SIZE, stevens says 24, but on linux (at least) it's 4 */
-/*   who knows if this will break anything */
 #define ETHHDR_SIZE 14 
 #define PPPHDR_SIZE 4 
 #define SLIPHDR_SIZE 16
@@ -19,11 +21,21 @@
 #define WORD_REGEX "((^%s\\W)|(\\W%s$)|(\\W%s\\W))"
 #define IP_ONLY "ip and ( %s)"
 
+
 char *get_filter(char **);
 void process(u_char *, struct pcap_pkthdr*, u_char *);
-int re_match_func(char *, int); 
-int blank_match_func(char *, int); 
 void dump(char *, int); 
 void clean_exit(int);
 void usage(int);
 void version(void);
+
+int re_match_func(char *, int); 
+int bin_match_func(char *, int);
+int blank_match_func(char *, int); 
+
+int strishex(char *);
+
+void print_time_absolute(struct pcap_pkthdr *);
+void print_time_diff(struct pcap_pkthdr *);
+
+void dump_delay(struct pcap_pkthdr *);
